/*
   FNT - Stream handler for socket connections
   Accept new connections

   Copyright (C) 2009 Free Software Foundation, Inc.

   Written by:
   Slava Zanko <slavazanko@gmail.com>


   This file is part of the FIDO New Technology.

   FIDO New Technology is free software: you can redistribute it and/or 
   modify it under the terms of the GNU General Public License as published 
   by the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifdef HAVE_CONFIG_H
#	include <config.h>
#endif

#include <string.h>

#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <errno.h>

#include "../include/global.h"
#include "../include/types.h"
#include "../include/error.h"
#include "../include/event.h"
#include "../include/stream.h"
#include "../include/mainloop.h"

#include "main.h"
#include "utils.h"

/*** global variables **************************************************/

/*** file scope macro definitions **************************************/

/*** file scope type declarations **************************************/

/*** file scope functions **********************************************/

/*** public functions **************************************************/

void
fnt_stream_socket_acceptCallback (GIOChannel * source, GIOCondition condition,
			   gpointer data)
{
    int new_fd, server_fd;
    __fnt_driver_data_t *driver_data, *new_driver_data;
    fnt_stream_t *new_stream, *listen_stream;
    socklen_t client_len = sizeof (struct sockaddr_in);
    struct sockaddr_in client_addr;

    GIOChannel *new_socket;
    GIOStatus status;
    GError *error = NULL;

    listen_stream = (fnt_stream_t *) data;
    driver_data = (__fnt_driver_data_t *) listen_stream->driver_data;


#ifdef G_OS_WIN32
    server_fd = g_io_channel_win32_get_fd (driver_data->socket);
#else
    server_fd = g_io_channel_unix_get_fd (driver_data->socket);
#endif


    memset (&client_addr, 0, sizeof (struct sockaddr_in));

    new_fd =
	accept (server_fd, (struct sockaddr *) &client_addr, &client_len);

    if (new_fd < 0)
    {
	fnt_error_push (FNT_MODULE_stream, FNT_STREAM_E_INTERNAL,
			_("Unable to accept new connection"));
	return;
    }

#ifdef G_OS_WIN32
    new_socket = g_io_channel_win32_new_socket (new_fd);
#else
    new_socket = g_io_channel_unix_new (new_fd);
#endif

    status = g_io_channel_set_flags (new_socket, G_IO_FLAG_NONBLOCK, &error);
    g_io_channel_set_encoding (new_socket, NULL, &error);
    g_io_channel_set_buffered (new_socket, FALSE);


    new_stream = fnt_stream_init ("socket");
    if (new_stream == NULL)
    {
	g_io_channel_shutdown (new_socket, TRUE, NULL);
	g_io_channel_unref (new_socket);
	return;
    }

    new_stream->mainloop = listen_stream->mainloop;
    new_stream->driver = listen_stream->driver;
    new_stream->state = FNT_STREAM_STATE_ok;

    new_driver_data = g_try_malloc0 (sizeof (__fnt_driver_data_t));
    if (new_driver_data == NULL)
    {
	fnt_error_push (FNT_MODULE_stream, FNT_STREAM_E_INTERNAL,
			_("Unable to allocate memory for '%s'"),
			"fnt_stream_t");
	g_io_channel_shutdown (new_socket, TRUE, NULL);
	g_io_channel_unref (new_socket);
	g_free (new_stream);
	return;
    }
    new_driver_data->socket = new_socket;
    new_stream->is_server = true;

    new_stream->driver_data = (void *) new_driver_data;

    new_stream->g_event_source =
	g_io_add_watch (new_driver_data->socket,
			G_IO_IN | G_IO_PRI | G_IO_HUP | G_IO_ERR | G_IO_NVAL,
			driver_data->callbackIo, new_stream);
    g_io_channel_unref (new_socket);


    fnt_event_raise ("stream", "onAccept", (void *) new_stream);
}
